"""
This code is used for Quantum Detector Tomography (QDOT) and not Quantum/Diagonal Detector Overlapping Tomography (QDOT/DDOT).
In the current version of QREM (2023.01.12) it only appears in tutorials in OldStuff
"""

import numpy as np
from qiskit import QuantumCircuit, QuantumRegister, ClassicalRegister
import itertools

from qrem.functions_qrem import povmtools
from qrem.functions_qrem.PyMaLi import GeneralTensorCalculator # old version: from PyMaLi import GeneralTensorCalculator


#ORGANIZE: All of these functions are almost exactly duplicated in QDT_pyquil. Delete one of them (carefully)
def get_list_of_lists_indices_qdt(qubits_indices, unitaries_amount):
    """
    Description:
    From list of qubit indices and number of unitaries, prepare set of names for circuits for tomography of
    detectors.

    # TODO FBM: Maybe GeneralTensor could be used for it?
    # TR: It now is -- check it out ;)

    :param qubits_indices: (list of ints) labels of qubits for QDT
    :param unitaries_amount: (int) number of unitaries you want to implement.

    :return: (list of lists of pairs of number [qubit_label, unitary_label]) of length
    unitaries_amount**len(qubits_indices)

    For example, let's assume we have three qubits [0,2,4] and we have 4 unitaries. For concreteness, if those unitaries
    are [id, X, H, SH] for creation of [|0>, |1>, |+x>, |+y>] states, the circuits will be:

    id x id x id,
    -------------
    id x id x X,
    -------------
    id x id x H,
    -------------
    id x id x SH,
    -------------

    id x X x id
    -------------
    id x X x X
    -------------
    id x X x H
    -------------
    id x X x SH
    -------------

    etc....

    SH x SH x id
    -------------
    SH x SH x X
    -------------
    SH x SH x H
    -------------
    SH x SH x SH

    If unitaries are ordered as [id, X, H, SH], the code will identify id: 0, X: 1, H: 2, SH:3, and it will return a
    list of lists. Each list correspond to single line in the above exemplary description of tensor product, e.g.,
    second list corresponds to line id x id x X. Each of such lists consists of len(qubits_indices) pairs, saying which
    gate shall be applied to which qubit. In this example, first list will have following elements:
    (0, 0), (2, 0), (4, 1), i_index.e., apply unitary number 0 to qubits 0,2 and apply unitary number 1 to qubit 4.
    """

    qubits_number = len(qubits_indices)

    # number unitaries from 0 to unitaries_amount-1
    unitaries_indices = range(unitaries_amount)

    # create Cartesian product of labels for qubits and unitaries
    qubits_and_unitaries = list(itertools.product(qubits_indices, unitaries_indices))

    # take lists corresponding to particular qubits
    single_qubit_unitaries = [qubits_and_unitaries[i * unitaries_amount:(i + 1) * unitaries_amount] for
                              i in
                              range(qubits_number)]

    # get all possible combinations of qubit+gate pairs
    list_of_gates = single_qubit_unitaries[0]
    for i in np.arange(1, qubits_number):
        list_of_gates = list(itertools.product(list_of_gates, single_qubit_unitaries[i]))

    # TODO FBM: this code could end here if I knew how to flatten list of itertools.product() s in a simple manner.
    #  Without next steps the list is a very nested tuple of tuples

    def flatten(container):
        # lame flattener
        for i in container:
            if isinstance(i, list) or isinstance(i, tuple):
                for j in flatten(i):
                    yield j
            else:
                yield i

    # flatten everything
    list_of_gates = list(flatten(list(list_of_gates)))

    # now put it back together in two steps
    single_circuit_description_length = qubits_number * 2

    # take sublists of length 2*qubits_number (one index for qubit and one index for unitary)
    almost_final_list = [
        list_of_gates[
        i * single_circuit_description_length:(i + 1) * single_circuit_description_length] for i in
        range(int(len(list_of_gates) / single_circuit_description_length))]

    final_list = []
    # create final list of lists for which entries are pairs (qubit_index, unitary_index)
    for l in almost_final_list:
        final_list.append([l[i * 2:(i + 1) * 2] for i in range(int(len(l) / 2))])

    return final_list


def QDT_circuits_qiskit(qubit_indices,
                        probe_kets,
                        number_of_repetitions=1,
                        qrs=None,
                        crs=None,
                        add_measurements=True):
    """From list of probe kets and qubit data return quantum circuits which will be implemented to perform
    Quantum Detector Tomography (QDT).

    :param probe_kets: (list of numpy arrays) the list of ket representations of qubit pure quantum states which are to
    be used in QDT. For multi-qubit QDT, the combinations of tensor products of name_appendix will be taken.
    :param qubit_indices: (list of ints) labels of qubits for QDT
    :param number_of_repetitions: (int) parameter specifying how many copies of whole QDT experiment should be created
    (for larger statistics collection or comparision of results)

    :return: (list of QuantumCircuit objects) of length len(probe_states)**(number_of_qubits)
    """

    qubit_indices = sorted(qubit_indices)  # Sort to ensure, that results can easily be interpreted.
    tomography_circuits = []
    unitaries = [povmtools.get_unitary_change_ket_qubit(ket) for ket in probe_kets]

    # create nice list with proper ordering of circuits. In first step, last qubit in list qubit_indices is iterated
    # while all the other are fixed. See function description for details.
    indices_for_circuits = get_list_of_lists_indices_qdt(qubit_indices, len(unitaries))
    if qrs is None:
        qrs = max(qubit_indices) + 1

    if crs is None:
        len(qubit_indices) # ORGANIZE - this does nothing, probably should be crs = len(qubit_indices). #TEST
    # outer loop is for copies of QDT experiment
    for number in range(number_of_repetitions):
        # inner loop goes through all circuits in QDT experiments and prepares them
        for index_for_set in range(len(indices_for_circuits)):
            # print(number,index_for_set)

            # index of set of qubits+unitaries for current step
            current_set = indices_for_circuits[index_for_set]

            # create quantum register
            qreg = QuantumRegister(qrs)

            creg = ClassicalRegister(crs)

            # create quantum circuit with nice names
            set_string = ''.join(['u' + str(st) for st in current_set[0]])
            qubits_string = ''.join(['q' + str(st) for st in qubit_indices])

            circuit = QuantumCircuit(qreg, creg,
                                     name="QDT-" + qubits_string + "-id-" + set_string + '-no-' + str(
                                         number))

            # get barrier to prevent compiler from making changes
            circuit.barrier()

            for qubit_unitary_pair_index in range(len(current_set)):
                # take qubit+unitary pair
                pair_now = current_set[qubit_unitary_pair_index]

                # take index of qubit and index of unitary
                q_now_index, u_now_index = pair_now[0], pair_now[1]

                # make sure that chosen quantum state is not one of the states in computational basis
                # TODO: this might not be necessary anymore, it's an old code, I had some problems long time ago with
                #  those guys because qiskit compiler went crazy if I defined identity or x gate using u3 unitary.
                if povmtools.check_if_projector_is_in_computational_basis(
                        povmtools.get_density_matrix(probe_kets[u_now_index])):
                    if povmtools.get_density_matrix(probe_kets[u_now_index])[0][0] == 1:
                        circuit.i(qreg[q_now_index])
                    elif povmtools.get_density_matrix(probe_kets[u_now_index])[1][1] == 1:
                        circuit.x(qreg[q_now_index])
                    else:
                        raise ValueError('error')
                else:
                    # get angles for single-qubit state change unitary
                    current_angles = povmtools.get_su2_parametrizing_angles(unitaries[u_now_index])

                    # implement unitary
                    circuit.u3(current_angles[0],
                               current_angles[1],
                               current_angles[2], qreg[q_now_index])

            # Add measurements
            if add_measurements:
                for i in range(len(qubit_indices)):
                    circuit.measure(qreg[qubit_indices[i]], creg[i])

            tomography_circuits.append(circuit)
    return tomography_circuits


def detector_tomography_circuits_rigetti(qubit_indices,
                                         probe_kets,
                                         shots=10 ** 5):
    """From list of probe kets and qubit data return quantum circuits which will be implemented to perform
    Quantum Detector Tomography (QDT).

    :param probe_kets: (list of numpy arrays) the list of ket representations of qubit pure quantum states which are to
    be used in QDT. For multi-qubit QDT, the combinations of tensor products of name_appendix will be taken.
    :param qubit_indices: (list of ints) labels of qubits for QDT
    :param number_of_repetitions: (int) parameter specifying how many copies of whole QDT experiment should be created
    (for larger statistics collection or comparision of results)

    :return: (list of QuantumCircuit objects) of length len(probe_states)**(number_of_qubits)
    """
    import pyquil as pyq
    # TODO: this is stupid fastly written code for 1q
    qubit_indices = sorted(qubit_indices)  # Sort to ensure, that results can easily be interpreted.
    tomography_circuits = []
    # unitaries = [povmtools.get_unitary_change_ket_qubit(ket) for ket in probe_kets]

    # create nice list with proper ordering of circuits. In first step, last qubit in list qubit_indices is iterated
    # while all the other are fixed. See function description for details.
    indices_for_circuits = get_list_of_lists_indices_qdt(qubit_indices, len(probe_kets))

    # inner loop goes through all circuits in QDT experiments and prepares them
    for index_for_set in range(len(indices_for_circuits)):

        # index of set of qubits+unitaries for current step
        current_set = indices_for_circuits[index_for_set]

        # create quantum register
        # qreg = QuantumRegister(quantum_register_size)
        # creg = ClassicalRegister(len(qubit_indices))

        # create quantum circuit with nice names
        set_string = ''.join(['u' + str(st) for st in current_set[0]])
        qubits_string = ''.join(['q' + str(st) for st in qubit_indices])

        #
        # circuit = QuantumCircuit(qreg, creg,
        #                          name="QDT-" + qubits_string + "-id-" + set_string + '-no-' + str(number))

        program = pyq.Program()
        ro = program.declare('ro',
                             memory_type='BIT',
                             memory_size=len(qubit_indices))
        # program += pyq.gates.RESET()
        # get barrier to prevent compiler from making changes
        # circuit.barrier()

        for qubit_unitary_pair_index in range(len(current_set)):
            # take qubit+unitary pair
            pair_now = current_set[qubit_unitary_pair_index]

            # take index of qubit and index of unitary
            q_now_index, u_now_index = pair_now[0], pair_now[1]

            if u_now_index == 0:
                program += pyq.gates.I(q_now_index)
            elif u_now_index == 1:
                program += pyq.gates.X(q_now_index)
            elif u_now_index == 2:
                program += pyq.gates.H(q_now_index)
            elif u_now_index == 3:
                program += pyq.gates.X(q_now_index)
                program += pyq.gates.H(q_now_index)
            elif u_now_index == 4:
                program += pyq.gates.H(q_now_index)
                program += pyq.gates.S(q_now_index)
            elif u_now_index == 5:
                program += pyq.gates.X(q_now_index)
                program += pyq.gates.H(q_now_index)
                program += pyq.gates.S(q_now_index)

        # Add measurements
        for i in range(len(qubit_indices)):
            program += pyq.gates.MEASURE(qubit_indices[i], ro[i])

        program.wrap_in_numshots_loop(shots)
        tomography_circuits.append(program)
    return tomography_circuits


def gtc_tensor_calculating_function(arguments: list):
    result = []

    for a in arguments:
        result.append(a)

    return result


def detector_tomography_circuits_pymali(qubit_indices, probe_kets):
    """
    Analogical method_name of the circuits preparing method_name utilizing pymali general tensor calculator.
    """

    qubit_indices = sorted(qubit_indices)  # Sort to ensure, that results can easily be interpreted.
    tomography_circuits = []
    unitaries = [povmtools.get_unitary_change_ket_qubit(ket) for ket in probe_kets]

    # create nice list with proper ordering of circuits. In first step, last qubit in list qubit_indices is iterated
    # while all the other are fixed. See function description for details.
    gtc = GeneralTensorCalculator.GeneralTensorCalculator(gtc_tensor_calculating_function)
    unitaries_lists_for_tensor_calculator = [unitaries.copy() for i in range(len(qubit_indices))]
    list_of_unitaries_sets = gtc.calculate_tensor_to_increasing_list(
        unitaries_lists_for_tensor_calculator)
    qrs = max(qubit_indices) + 1

    # inner loop goes through all circuits in QDT experiments and prepares them
    for unitaries_set in list_of_unitaries_sets:

        qreg = QuantumRegister(qrs)
        creg = ClassicalRegister(len(qubit_indices))

        circuit = QuantumCircuit(qreg, creg)
        circuit.barrier()  # To prevent compiler from making changes.

        for j in range(len(unitaries_set)):

            current_angles = povmtools.get_su2_parametrizing_angles(unitaries_set[j])

            # TODO TR: I believe there may be more "special" cases.
            #  If so, then this should be placed in other method_name
            #  or in get_su2_ ... method_name.
            if current_angles[0] == 'id':
                circuit.i(qreg[qubit_indices[j]])
                continue
            if current_angles[0] == 'x':
                circuit.x(qreg[qubit_indices[j]])
                continue

            # implement unitary
            circuit.u3(current_angles[0],
                       current_angles[1],
                       current_angles[2], qreg[qubit_indices[j]])

        # Add measurements
        for i in range(len(qubit_indices)):
            circuit.measure(qreg[qubit_indices[i]], creg[i])

        tomography_circuits.append(circuit)
    return tomography_circuits



"""
@author: Jan Tuziemski, Filip Maciejewski, Joanna Majsak, Oskar Słowik, Marcin Kotowski, Katarzyna Kowalczyk-Murynka, Paweł Przewłocki, Piotr Podziemski, Michał Oszmaniec
@contact: filip.b.maciejewski@gmail.com
"""
from typing import Optional, Dict, Tuple

import numpy as np
import picos
from tqdm import tqdm

from qrem.functions_qrem import ancillary_functions as anf
from qrem.noise_mitigation.expected_values.CRS.scipy_optimizers_wrappers import \
    wrapper_scipy_optimize as scop_wrapper

import qrem.common.math as qrem_math
from qrem.common.printer import qprint, qprint_array

def get_symmetrization_matrix_1q(stochastic_matrix: np.ndarray):
    p10, p01 = stochastic_matrix[1, 0], stochastic_matrix[0, 1]

    p, q = p10, p01

    delta = q - p

    b = np.max([0, -delta / (1 - delta)])

    dp, dm = 1 + delta, 1 - delta

    a = (b * dm + delta) / dp

    Ai = np.array([[1 - a, b], [a, 1 - b]])

    if qrem_math.is_matrix_stochastic(Ai):
        return Ai
    else:
        raise ValueError('Something is wrong, symmetrization matrix not stochastic.')


def get_equalization_matrix_1q(symmetrized_matrix,
                               max_probability):
    pi = symmetrized_matrix[0, 1]
    bi = (max_probability - pi) / (1 - 2 * pi)

    Bi = np.array([[1 - bi, bi], [bi, 1 - bi]])

    if qrem_math.is_matrix_stochastic(Bi):
        return Bi
    else:
        qprint_array(Bi)
        raise ValueError('Something is wrong, equalization matrix not stochastic.')


def find_general_symmetrizer(stochastic_matrix,
                             equalized_sucess_probability=None):
    dimension = stochastic_matrix.shape[0]

    problem = picos.Problem()
    symmetrizer = picos.RealVariable(name='S',
                                     shape=(dimension, dimension))
    if equalized_sucess_probability is None:
        p_succ = picos.RealVariable(name='psucc')
        maximization = True

    else:
        p_succ = equalized_sucess_probability
        maximization = False

    symmetrized_matrix = symmetrizer * stochastic_matrix

    for i in range(dimension):
        problem.add_constraint(picos.sum(symmetrizer[:, i]) == 1)

    for i in range(dimension):
        for j in range(dimension):
            problem.add_constraint(symmetrizer[i, j] >= 0)
    for i in range(dimension):
        problem.add_constraint(symmetrized_matrix[i, i] == p_succ)

    for i in range(dimension):
        for j in range(i + 1, dimension):
            problem.add_constraint(symmetrized_matrix[i, j] == (1 - p_succ) / (dimension - 1))
            problem.add_constraint(symmetrized_matrix[j, i] == (1 - p_succ) / (dimension - 1))

    if maximization:
        problem.set_objective('max',
                              expression=p_succ)
    else:
        problem.set_objective('find')
    problem.solve(solver='mosek')

    symmetrizer = np.array(symmetrizer.value_as_matrix)
    return symmetrizer


def find_stochastic_correction(stochastic_matrix_1q):
    dimension = 2

    problem = picos.Problem()
    improver = picos.RealVariable(name='S',
                                  shape=(dimension, dimension))

    p_succ = picos.RealVariable(name='psucc')
    improved_matrix = improver * stochastic_matrix_1q

    for i in range(dimension):
        problem.add_constraint(picos.sum(improver[:, i]) == 1)

    for i in range(dimension):
        for j in range(dimension):
            problem.add_constraint(improver[i, j] >= 0)
    for i in range(dimension):
        problem.add_constraint(improved_matrix[i, i] == p_succ)

    for i in range(dimension):
        for j in range(i + 1, dimension):
            problem.add_constraint(improved_matrix[i, j] == (1 - p_succ) / (dimension - 1))
            problem.add_constraint(improved_matrix[j, i] == (1 - p_succ) / (dimension - 1))

    problem.set_objective('max',
                          expression=p_succ)
    problem.solve(solver='mosek')

    improver = np.array(improver.value_as_matrix)

    return improver


def get_max_error_probability(symmetrized_noise_matrices):
    return np.max([noise_matrix[0, 1] for noise_matrix in symmetrized_noise_matrices])


def get_1q_lambda(p, q=None):
    if q is None:
        q = p

    return np.array([[1 - p, q],
                     [p, 1 - q]])


def get_tensored_basis_birkhoff(number_of_qubits):
    basis_1q = [np.eye(2, dtype=float),
                np.array([[0, 1], [1, 0]])]

    from qrem.functions_qrem.PyMaLi import GeneralTensorCalculator # old version: from PyMaLi import GeneralTensorCalculator 

    gtc = GeneralTensorCalculator()
    basis_birkhoff = gtc.calculate_tensor_to_increasing_list(
        objects=[basis_1q for _ in range(number_of_qubits)])

    return basis_birkhoff


def move_matrix_to_tensored_birkhoff_polytope_with_postprocessing(
        initial_stochastic_map,
        return_mover=True):
    dimension = initial_stochastic_map.shape[0]
    number_of_qubits = int(np.log2(dimension))

    problem = picos.Problem()
    mover_to_Birkhoff = picos.RealVariable(name='mover_to_Birkhoff',
                                           shape=(dimension, dimension),
                                           lower=0, upper=1)
    product = mover_to_Birkhoff * initial_stochastic_map

    for column_index in range(dimension):
        problem.add_constraint(picos.sum(mover_to_Birkhoff[:, column_index]) == 1)

    tensored_basis = get_tensored_basis_birkhoff(number_of_qubits=number_of_qubits)
    product_coeffs = picos.RealVariable('product_coeffs',
                                        shape=(dimension),
                                        lower=0,
                                        upper=1)
    problem.add_constraint(picos.sum(product_coeffs) == 1)

    tensored_birkhoff_element = 0
    for i in range(dimension):
        tensored_birkhoff_element += product_coeffs[i] * tensored_basis[i]

    problem.add_constraint(tensored_birkhoff_element == product)

    cost_function = product_coeffs[0]
    problem.set_objective(direction='max',
                          expression=cost_function
                          )
    problem.solve(solver='mosek')
    if return_mover:
        return np.array(mover_to_Birkhoff.value_as_matrix), np.array(product.value_as_matrix)
    else:
        return cost_function.coherences_values


def move_matrix_to_tensored_birkhoff_polytope_with_preprocessing(
        initial_stochastic_map,
        return_mover=True):
    dimension = initial_stochastic_map.shape[0]
    number_of_qubits = int(np.log2(dimension))

    tensored_basis = get_tensored_basis_birkhoff(number_of_qubits=number_of_qubits)

    problem = picos.Problem()

    movers_to_Birkhoff = [
        picos.RealVariable(name='mover_to_Birkhoff_%s' % i,
                           shape=(dimension, dimension),
                           lower=0, upper=1) for i in range(dimension)]

    rs = [picos.RealVariable(name='r_%s' % i, shape=1, lower=0, upper=1) for i in range(dimension)]

    problem.add_constraint(picos.sum(rs) == 1)
    product = 0
    for i in range(dimension):
        Lami, ri = movers_to_Birkhoff[i], rs[i]
        permuted_initital_lambda = initial_stochastic_map @ tensored_basis[i]
        product += Lami * permuted_initital_lambda
        for k in range(dimension):
            problem.add_constraint(picos.sum(Lami[:, k]) == ri)

    product_coeffs = picos.RealVariable('product_coeffs',
                                        shape=(dimension),
                                        lower=0,
                                        upper=1)
    problem.add_constraint(picos.sum(product_coeffs) == 1)

    tensored_birkhoff_element = 0
    for i in range(dimension):
        tensored_birkhoff_element += product_coeffs[i] * tensored_basis[i]

    problem.add_constraint(tensored_birkhoff_element == product)

    cost_function = product_coeffs[0]
    problem.set_objective(direction='max',
                          expression=cost_function
                          )
    problem.solve(solver='mosek')

    if return_mover:
        return [np.array(mover_to_Birkhoff.value_as_matrix) for mover_to_Birkhoff in
                movers_to_Birkhoff], [ri.value for ri in rs], np.array(product.value_as_matrix)
    else:
        return cost_function.coherences_values


def find_separable_decomposition_vectorized(params,
                                            element_of_tensored_birkhoff_polytope,
                                            minimize=True,
                                            tensored_vector=None,
                                            return_decoupler=False):
    dimension = element_of_tensored_birkhoff_polytope.shape[0]
    if tensored_vector is None:
        tensored_vector = 1
        for p in params:
            tensored_vector = np.kron(tensored_vector, np.array([1 - p, p]))

    problem = picos.Problem()
    coeffs_vector = picos.RealVariable(name='coeffs_vector', shape=(dimension, 1))
    product = element_of_tensored_birkhoff_polytope * coeffs_vector

    for column_index in range(dimension):
        problem.add_constraint(coeffs_vector[column_index] >= 0)
    problem.add_constraint(picos.sum(coeffs_vector) == 1)

    if minimize:
        t = picos.Norm(product - tensored_vector, p=2, q=2)
        problem.set_objective(direction='min',
                              expression=t
                              )
        problem.solve(solver='mosek')

        if return_decoupler:
            number_of_qubits = int(np.log2(dimension))
            tensored_basis = get_tensored_basis_birkhoff(number_of_qubits=number_of_qubits)
            decoupler = 0
            for i in range(dimension):
                decoupler += tensored_basis[i] * coeffs_vector[i].coherences_values

            return decoupler
        else:
            return t.value


    else:
        problem.add_constraint(product == tensored_vector)
        problem.set_objective(direction='find',
                              )

        try:
            problem.solve(solver='mosek')
            if problem.status == 'optimal':
                if return_decoupler:
                    number_of_qubits = int(np.log2(dimension))
                    tensored_basis = get_tensored_basis_birkhoff(number_of_qubits=number_of_qubits)
                    decoupler = 0
                    for i in range(dimension):
                        decoupler += tensored_basis[i] * coeffs_vector[i].coherences_values

                    return decoupler
                else:
                    return True
            else:
                return False
        except:
            return False


def find_separable_decomposition_not_vectorized_with_postprocessing(params,
                                                                    element_of_tensored_birkhoff_polytope,
                                                                    tensored_matrix=None,
                                                                    return_decoupler=False,
                                                                    symmetric=True):
    dimension = element_of_tensored_birkhoff_polytope.shape[0]
    number_of_qubits = int(np.log2(dimension))
    tensored_basis = get_tensored_basis_birkhoff(number_of_qubits=number_of_qubits)

    if tensored_matrix is None:
        tensored_matrix = 1

        if symmetric:
            for p in params:
                tensored_matrix = np.kron(tensored_matrix, np.array([[1 - p, p],
                                                                     [p, 1 - p]]))

        else:
            for i in range(int(len(params) / 2)):
                tensored_matrix = np.kron(tensored_matrix,
                                          np.array([[1 - params[2 * i], params[2 * i + 1]],
                                                    [params[2 * i], 1 - params[2 * i + 1]]]))

    problem = picos.Problem()

    decoupler = picos.RealVariable(name='decoupler', shape=(dimension, dimension),
                                   lower=0, upper=1)

    product = decoupler * element_of_tensored_birkhoff_polytope
    for i in range(dimension):
        problem.add_constraint(picos.sum(decoupler[:, i]) == 1)

    coeffs_vector = picos.RealVariable(name='coeffs_vector', shape=(dimension, 1), lower=0, upper=1)
    problem.add_constraint(picos.sum(coeffs_vector) == 1)
    problem.add_constraint(
        product == picos.sum([coeffs_vector[i] * tensored_basis[i] for i in range(dimension)]))

    t = picos.Norm(product - tensored_matrix, p=2, q=2)
    problem.set_objective(direction='min',
                          expression=t
                          )
    problem.solve(solver='mosek')

    if return_decoupler:
        return np.array(decoupler.value_as_matrix), np.array(product.value_as_matrix)

    else:
        return t.value


def find_separable_decomposition_not_vectorized_with_preprocessing(params,
                                                                   element_of_tensored_birkhoff_polytope,
                                                                   tensored_matrix=None,
                                                                   return_decoupler=False,
                                                                   symmetric=True):
    dimension = element_of_tensored_birkhoff_polytope.shape[0]
    number_of_qubits = int(np.log2(dimension))
    tensored_basis = get_tensored_basis_birkhoff(number_of_qubits=number_of_qubits)

    if tensored_matrix is None:
        tensored_matrix = 1
        if symmetric:
            for p in params:
                tensored_matrix = np.kron(tensored_matrix, np.array([[1 - p, p],
                                                                     [p, 1 - p]]))

        else:
            for i in range(int(len(params) / 2)):
                tensored_matrix = np.kron(tensored_matrix,
                                          np.array([[1 - params[2 * i], params[2 * i + 1]],
                                                    [params[2 * i], 1 - params[2 * i + 1]]]))

    problem = picos.Problem()

    pre_post = [
        picos.RealVariable(name='PP%s' % i, shape=(dimension, dimension),
                           lower=0, upper=1) for i in range(dimension)]
    rs = [picos.RealVariable(name='r_%s' % i, shape=1, lower=0, upper=1) for i in range(dimension)]

    problem.add_constraint(picos.sum(rs) == 1)
    product = 0
    for i in range(dimension):
        Lami, ri = pre_post[i], rs[i]
        permuted_lambda = element_of_tensored_birkhoff_polytope @ tensored_basis[i]
        product += Lami * permuted_lambda
        for k in range(dimension):
            problem.add_constraint(picos.sum(Lami[:, k]) == ri)

    # ps = [picos.RealVariable(name='p_%s' % i, shape=1, lower=0, upper=1) for i in range(dimension)]
    coeffs_vector = picos.RealVariable(name='coeffs_vector', shape=(dimension, 1), lower=0, upper=1)
    problem.add_constraint(picos.sum(coeffs_vector) == 1)
    problem.add_constraint(
        product == picos.sum([coeffs_vector[i] * tensored_basis[i] for i in range(dimension)]))
    # tensored_matrix

    t = picos.Norm(product - tensored_matrix, p=2, q=2)
    problem.set_objective(direction='min',
                          expression=t
                          )
    problem.solve(solver='mosek')

    if return_decoupler:
        postprocessing_matirces = [np.array(PP.value_as_matrix) for PP in
                                   pre_post]
        preprocessing_distribution = [ri.value for ri in rs]

        decoupled_matrix = np.array(product.value_as_matrix)

        return postprocessing_matirces, preprocessing_distribution, decoupled_matrix

    else:
        return t.value


def find_separable_decomposition_general(params,
                                         element_of_tensored_birkhoff_polytope,
                                         only_postprocessing: bool,
                                         symmetric: bool,
                                         tensored_matrix=None,
                                         return_decoupler=False,
                                         ):
    if only_postprocessing:
        return find_separable_decomposition_not_vectorized_with_postprocessing(params=params,
                                                                               element_of_tensored_birkhoff_polytope=element_of_tensored_birkhoff_polytope,
                                                                               tensored_matrix=tensored_matrix,
                                                                               return_decoupler=return_decoupler,
                                                                               symmetric=symmetric)
    else:
        return find_separable_decomposition_not_vectorized_with_preprocessing(params=params,
                                                                              element_of_tensored_birkhoff_polytope=element_of_tensored_birkhoff_polytope,
                                                                              tensored_matrix=tensored_matrix,
                                                                              return_decoupler=return_decoupler,
                                                                              symmetric=symmetric)


def cost_function_vectorized(params,
                             element_of_tensored_birkhoff_polytope,
                             tensored_vector=None
                             ):
    distance = find_separable_decomposition_vectorized(params=params,
                                                       element_of_tensored_birkhoff_polytope=element_of_tensored_birkhoff_polytope,
                                                       minimize=True,
                                                       tensored_vector=tensored_vector)

    cost_function = 0
    cost_function += np.sum(params)

    # for i in range(len(params)):
    #     for j in range(i+1,len(params)):
    #         pi, pj = params[i], params[j]
    #         cost_function+=1/(1-2*pi)*1/(1-2*pj)

    cost_function += 10 ** 6 * distance

    return cost_function


def cost_function_not_vectorized(params,
                                 element_of_tensored_birkhoff_polytope,
                                 symmetric,
                                 only_postprocessing,
                                 tensored_matrix=None,
                                 ):
    distance = find_separable_decomposition_general(params=params,
                                                    element_of_tensored_birkhoff_polytope=element_of_tensored_birkhoff_polytope,
                                                    only_postprocessing=only_postprocessing,
                                                    tensored_matrix=tensored_matrix,
                                                    symmetric=symmetric)
    #
    # distance = find_separable_decomposition_not_vectorized_with_preprocessing(params=params,
    #                                                                           element_of_tensored_birkhoff_polytope=element_of_tensored_birkhoff_polytope,
    #                                                                           tensored_matrix=tensored_matrix,
    #                                                                           symmetric=symmetric)

    cost_function = 0
    cost_function += np.sum(params)

    # for i in range(len(params)):
    #     for j in range(i+1,len(params)):
    #         pi, pj = params[i], params[j]
    #         cost_function+=1/(1-2*pi)*1/(1-2*pj)

    cost_function += 10 ** 7 * distance

    return cost_function


def find_optimal_tensored_decomposition_blackbox(element_of_tensored_birkhoff_polytope,
                                                 symmetric,
                                                 only_postprocessing,
                                                 printing=True,
                                                 max_error_probability=0.45,
                                                 ):
    # print('hejka')
    dimension = element_of_tensored_birkhoff_polytope.shape[0]
    number_of_qubits = int(np.log2(dimension))

    if symmetric:
        number_of_parameters = number_of_qubits
    else:
        number_of_parameters = 2 * number_of_qubits

    bounds_list = [(0, max_error_probability) for _ in range(number_of_parameters)]

    optimizer_inital = 'differential_evolution'
    optimizer_final = 'COBYLA'

    maxiter_initial = 10 ** 2
    maxiter_final = 10 ** 2
    # printing = True

    basinhopping_options_initial = {'niter': 10,
                                    'T': 0.1,
                                    'stepsize': 0.1,
                                    'disp': printing}
    params_starting = [max_error_probability for _ in
                       range(number_of_parameters)]

    basinhopping_options_final = {'niter': 10,
                                  'T': 0.01,
                                  'stepsize': 0.01,
                                  'disp': printing,
                                  'interval':3}

    differential_evolution_options_initial = {'maxiter': maxiter_initial,
                                              'popsize': 48,
                                              'workers': -1,
                                              'updating': 'deferred',
                                              'disp': printing}

    initial_kwargs = {'basinhopping_kwargs': basinhopping_options_initial,
                      'differential_evolution_kwargs': differential_evolution_options_initial}

    if optimizer_inital.upper() in ['DIFFERENTIAL_EVOLUTION']:
        basinhopping = False
    else:
        basinhopping = True

    if only_postprocessing:
        if symmetric:
            cost_function_now = cost_function_vectorized
            additional_arguments = (element_of_tensored_birkhoff_polytope,)
        else:
            cost_function_now = cost_function_not_vectorized
            additional_arguments = (element_of_tensored_birkhoff_polytope,
                                    symmetric,
                                    only_postprocessing
                                    )
    else:
        cost_function_now = cost_function_not_vectorized
        additional_arguments = (element_of_tensored_birkhoff_polytope,
                                symmetric,
                                only_postprocessing
                                )

    res = scop_wrapper(initial_parameters=params_starting,
                       target_function=cost_function_now,
                       optimizer_name=optimizer_inital,
                       basinhopping=basinhopping,
                       additional_arguments=additional_arguments,
                       bounds_list=bounds_list,
                       options={'maxiter': maxiter_initial},
                       **initial_kwargs
                       )



    best_xs = res.x
    best_funopt = res.fun
    if printing:
        qprint("initial optimal parameters:", best_xs)
        qprint("initial optimal function value:", best_funopt)

    # print(params_starting_next)
    res = scop_wrapper(initial_parameters=best_xs,
                       target_function=cost_function_now,
                       optimizer_name=optimizer_final,
                       basinhopping=True,
                       additional_arguments=additional_arguments,
                       bounds_list=bounds_list,
                       options={'maxiter': maxiter_final},
                       basinhopping_kwargs=basinhopping_options_final
                       )

    xopt = res.x
    funopt = res.fun

    if funopt < best_funopt:
        best_xs = xopt
        best_funopt = funopt

    if printing:
        print()
        qprint("final optimal parameters:", best_xs)
        qprint("final optimal function value:", best_funopt)
        print()

    return best_xs


def find_optimal_decoupler_postprocessing_symmetric(stochastic_matrix: np.ndarray,
                                                    printing: Optional[bool] = False,
                                                    ):
    mover, moved_matrix = move_matrix_to_tensored_birkhoff_polytope_with_postprocessing(
        initial_stochastic_map=stochastic_matrix,
        return_mover=True)
    # print('done')
    #
    error_probabilities = find_optimal_tensored_decomposition_blackbox(
        element_of_tensored_birkhoff_polytope=moved_matrix,
        printing=printing,
        symmetric=True,
        only_postprocessing=True)
    # print('done2')

    decoupler = find_separable_decomposition_vectorized(params=error_probabilities,
                                                        element_of_tensored_birkhoff_polytope=moved_matrix,
                                                        minimize=True,
                                                        return_decoupler=True
                                                        )

    symmetrizer = decoupler @ mover


    # symm = symmetrizer@stochastic_matrix
    # kronk = 1
    #
    # for qi in error_probabilities:
    #     kronk = np.kron(kronk,get_1q_lambda(qi))
    #
    # qprint_array(symm)
    # qprint_array(kronk)
    #
    #
    # raise KeyboardInterrupt

    decoupled_matrix = symmetrizer@stochastic_matrix

    return symmetrizer, error_probabilities, decoupled_matrix


def find_optimal_decoupler_general(stochastic_matrix: np.ndarray,
                                   symmetric: bool,
                                   only_postprocessing: bool,
                                   printing: Optional[bool] = False,
                                   ):
    # if symmetric and only_postprocessing:
    #

    if only_postprocessing:
        if symmetric:
            decoupler, error_probabilities, decoupled_matrix = find_optimal_decoupler_postprocessing_symmetric(stochastic_matrix=stochastic_matrix,
                                                                   printing=printing)

            return decoupler, error_probabilities, decoupled_matrix


        else:
            mover, moved_matrix = move_matrix_to_tensored_birkhoff_polytope_with_postprocessing(
                initial_stochastic_map=stochastic_matrix,
                return_mover=True)

    else:
        postprocessing_matrices, preprocessing_distribution, moved_matrix = move_matrix_to_tensored_birkhoff_polytope_with_preprocessing(
            initial_stochastic_map=stochastic_matrix,
            return_mover=True)

    error_probabilities = find_optimal_tensored_decomposition_blackbox(
        element_of_tensored_birkhoff_polytope=moved_matrix,
        printing=printing,
        symmetric=symmetric,
        only_postprocessing=only_postprocessing)

    if only_postprocessing:
        decoupler, decoupled_matrix = find_separable_decomposition_not_vectorized_with_postprocessing(
            params=error_probabilities,
            element_of_tensored_birkhoff_polytope=moved_matrix,
            return_decoupler=True,
            symmetric=symmetric)
        #TODO FBM: add error probabilities
        return decoupler@mover, None, decoupled_matrix

    else:
        postprocessing_matrices, preprocessing_distribution, decoupled_matrix = find_separable_decomposition_not_vectorized_with_preprocessing(
            params=error_probabilities,
            element_of_tensored_birkhoff_polytope=moved_matrix,
            return_decoupler=True,
            symmetric=symmetric)

        # print(preprocessing_distribution)
        return preprocessing_distribution, postprocessing_matrices, error_probabilities, decoupled_matrix






def get_CRS_dictionaries_from_noise_matrices_only_preprocessing(local_noise_matrices:Dict[Tuple[int], np.ndarray],
                                                                printing:Optional[bool]=False,
                                                                print_progress_bar:Optional[bool]=True):
    error_probabilitities_dictionary = {}
    symmetrizers_dictionary = {}

    for cluster, local_map in tqdm(local_noise_matrices.items(), disable= not print_progress_bar):
        # qprint_array(local_map)
        if len(cluster) == 1:
            equalizer = find_general_symmetrizer(stochastic_matrix=local_map
                                                         )
            equalized_matrix = equalizer @ local_map
            error_probability = equalized_matrix[0, 1]

            # errors_probabilities = [error_probability]
            error_probabilitities_dictionary[cluster] = error_probability
        else:
            equalizer, errors_probabilities = find_optimal_decoupler_general(
                stochastic_matrix=local_map,
                symmetric=True,
            only_postprocessing=True,
            printing=printing)
            for index_qubit in range(len(errors_probabilities)):
                error_probabilitities_dictionary[(cluster[index_qubit],)] = errors_probabilities[
                    index_qubit]

        symmetrizers_dictionary[cluster] = equalizer



    return symmetrizers_dictionary, error_probabilitities_dictionary

def get_CRS_dictionaries_from_noise_matrices(local_noise_matrices:Dict[Tuple[int], np.ndarray],
                                                                printing:Optional[bool]=False,
                                                                print_progress_bar:Optional[bool]=True):
    error_probabilitities_dictionary = {}
    symmetrizers_dictionary = {}

    for cluster, local_map in tqdm(local_noise_matrices.items(), disable= not print_progress_bar):
        # qprint_array(local_map)
        if len(cluster) == 1:
            equalizer = find_general_symmetrizer(stochastic_matrix=local_map
                                                         )
            equalized_matrix = equalizer @ local_map
            error_probability = equalized_matrix[0, 1]

            # errors_probabilities = [error_probability]
            error_probabilitities_dictionary[cluster] = error_probability
        else:
            equalizer, errors_probabilities = find_optimal_decoupler_general(
                stochastic_matrix=local_map,
                symmetric=True,
            only_postprocessing=False,
            printing=printing)
            for index_qubit in range(len(errors_probabilities)):
                error_probabilitities_dictionary[(cluster[index_qubit],)] = errors_probabilities[
                    index_qubit]

        symmetrizers_dictionary[cluster] = equalizer



    return symmetrizers_dictionary, error_probabilitities_dictionary




def get_error_mitigation_multipliers_2_local(error_probabilities_dictionary,
                                             number_of_qubits):
    multipliers_mitigation = {}
    for i in range(number_of_qubits):
        for j in range(i + 1, number_of_qubits):
            multiplier = 1
            for qi in (i, j):
                multiplier /= (1 - 2 * error_probabilities_dictionary[(qi,)])
            multipliers_mitigation[(i, j)] = multiplier
    for i in range(number_of_qubits):
        multipliers_mitigation[(i,)] = 1 / (1 - 2 * error_probabilities_dictionary[(i,)])
    return multipliers_mitigation


import operator
from functools import reduce
from qrem.types import CNNoiseModel
from typing import Dict, Tuple, List
import numpy as np
from qrem.common import probability, math
from qrem.functions_qrem import povmtools 
from qrem.functions_qrem import functions_data_analysis as fda





def mitigate_marginal(marginal: Tuple, results_dictionary: Dict[str, Dict[str, int]], noise_model:type[CNNoiseModel],ensure_proper_probability_distribution:bool = False ) ->np.array:
        
    """
    Function performs mitigation of a single marginal

    Parameters
    ----------
    marginal : tuple
        A tuple of qubits indices specifying the marginal e.g. (0,1)   
    result_dictionary:
        A nested dictionary with experimental results to be mitigated. Key corresponds to bitstring encoding input state, values to a dictionary with results bitstrings and counts  
        
    noise_model : object of CNNoiseModel class
        an object of CNNoiseModel class

    ensure_proper_probability_distribution: bool
        a boolean specifying whether mitigated pseudo probability distribution should be projected onto proper probability distribution, by default set to False 


    Returns
    -------
    mitigated_marginal_probability_distribution
        a numpy array with marginal (pseudo) marginal probability distribution 


    """

    #JT: a list of clusters that are involved in marginal is created
    clusters_in_marginal_list = get_clusters_in_marginal_list(marginal=marginal, noise_model=noise_model)
    
    #creates a list of qubit indices involved in a marginal
    unordered_qubits_in_marginal_list = list(reduce(operator.concat, clusters_in_marginal_list))
    
    #creates an inverse noise matrix for the marginal 
    clusters_inverse_noise_matrix = get_marginal_inverse_noise_matrix(noise_model=noise_model, clusters_in_marginal_list=clusters_in_marginal_list)

    #qubits list is sorted

    unordered_qubits_in_marginal_list.sort()

    #computes extended marginal
    clusters_marginal_counts=probability.compute_marginals(results_dictionary=results_dictionary,subsets_list=[tuple(unordered_qubits_in_marginal_list)])

    #we are interested in the first key only hence the construction nex(iter()) TODO: this does not work, start from here 
    mitigated_clusters_marginal_probability_distribution = fda.convert_subset_counts_dictionary_to_probability_distribution(clusters_marginal_counts)
    
    #performs mitigation on an extended marginal  
    mitigated_clusters_marginal_probability_distribution = clusters_inverse_noise_matrix.dot(mitigated_clusters_marginal_probability_distribution)

    if ensure_proper_probability_distribution:
        if not probability.is_valid_probability_vector(mitigated_clusters_marginal_probability_distribution):
           mitigated_clusters_marginal_probability_distribution = povmtools.find_closest_prob_vector_l2(mitigated_clusters_marginal_probability_distribution)

        

    #returns marginalization of an  extended marginal

    mitigated_marginal_probability_distribution = probability.compute_marginal_of_probability_distribution(mitigated_clusters_marginal_probability_distribution,[unordered_qubits_in_marginal_list.index(qubit) for qubit in marginal])

    return mitigated_marginal_probability_distribution.flatten()

def mitigate_marginals(marginals_list: List[Tuple], results_dictionary: Dict[str, Dict[str, int]], noise_model: type[CNNoiseModel] ,ensure_proper_probability_distribution = False )->Dict[str, Dict[tuple[int], np.array]]:
    """Function performs mitigation for marginals list

    Parameters
    ----------
    marginals_list : List[tuple]
        A list of tuples encoding marginals e.g. [(0,1),(4,6)]   

    result_dictionary:
        A nested dictionary with experimental results to be mitigated. Key corresponds to bitstring encoding input state, values to a dictionary with results bitstrings and counts  
        
    noise_model : object of CNNoiseModel class
        an object of CNNoiseModel class

    ensure_proper_probability_distribution: bool
        a boolean specifying whether mitigated pseudo probability distribution should be projected onto proper probability distribution, by default set to False 


    Returns
    -------

        a nested dictionary {input bitstring : {(subset_tuple): marginal_probability_distributution}}, where input bitstring encodes input setting, subset_tuple encodes qubits in the marginal,
        and marginal_probability_distributution is a (pseudo) probability distribution 
    """
    #JT this dictionary 
    mitigated_marginals_dictionary ={}
   
    for marginal in marginals_list:
        mitigated_marginals_dictionary[marginal] = mitigate_marginal(marginal=marginal,results_dictionary=results_dictionary,noise_model=noise_model,ensure_proper_probability_distribution=ensure_proper_probability_distribution)
   
    return {next(iter(results_dictionary)): mitigated_marginals_dictionary}
 
def get_marginal_inverse_noise_matrix(noise_model : type[CNNoiseModel], clusters_in_marginal_list:List[Tuple])-> np.array:
    """Function computes an inverse noise matrix for specified clusters  

    Parameters
    ----------
    noise_model : object of CNNoiseModel class
        An object of CNNoiseModel class 

    clusters_in_marginal_list:
        A list of tuples with clusters involved in the marginal   
        

    Returns
    -------

        An inverse noise matrix for qubits specified in clusters_in_marginal_list, qubits are sorted in ascending order 
        (e.g. for a clusters_in_marginal_list =[(0,4),(1,8)], indices of the inverse noise matrix indices correspond to qubits in the order (0,1,4,8) )
    
    """   
    
    unordered_qubits_in_marginal_list = list(reduce(operator.concat, clusters_in_marginal_list))

    
    marginal_inverse_noise_matrix = np.array([1])
    
    for cluster in clusters_in_marginal_list:
        marginal_inverse_noise_matrix = np.kron(marginal_inverse_noise_matrix, noise_model.inverse_noise_matrices[cluster])
  
    return math.permute_composite_matrix(qubits_list=unordered_qubits_in_marginal_list,noise_matrix=marginal_inverse_noise_matrix)

def get_clusters_in_marginal_list(marginal : Tuple[int], noise_model : type[CNNoiseModel]) -> List[Tuple]:
    """Function creates a list of clusters that are involved in a marginal

    Parameters
    ----------
    marginal : tuple
        A tuple specifying marginal 

    noise_model : object of CNNoiseModel class
        An object of CNNoiseModel class  
        

    Returns
    -------

    clusters_in_marginal_list
        A list of tuples involved in the input marginal 
        
    
    """  
    clusters_in_marginal_list=[]
    
    
    for qubit in marginal:
        cluster = noise_model.qubit_in_cluster[(qubit,)]
        if cluster not in clusters_in_marginal_list:
            clusters_in_marginal_list.append(noise_model.qubit_in_cluster[(qubit,)])
    return clusters_in_marginal_list



##############################################
######For tests purposes######################
#############################################

def mitigate_marginal_product(marginal: Tuple, results_dictionary: Dict[str, Dict[str, int]], noise_model:type[CNNoiseModel],ensure_proper_probability_distribution:bool = False ) ->np.array:
    
    #JT: a list of clusters that are involved in marginal is created
    clusters_in_marginal_list = get_clusters_in_marginal_list(marginal=marginal, noise_model=noise_model)

    final_marginal = np.array([1])

    for cluster in clusters_in_marginal_list:

        clusters_inverse_noise_matrix = get_marginal_inverse_noise_matrix(noise_model=noise_model, clusters_in_marginal_list=[cluster])

        clusters_marginal_counts=probability.compute_marginals(results_dictionary=results_dictionary,subsets_list=[cluster])
        
        mitigated_clusters_marginal_probability_distribution = fda.convert_subset_counts_dictionary_to_probability_distribution(clusters_marginal_counts)

        mitigated_clusters_marginal_probability_distribution = clusters_inverse_noise_matrix.dot(mitigated_clusters_marginal_probability_distribution)

        final_marginal = np.kron(final_marginal,mitigated_clusters_marginal_probability_distribution)
    



    
    #creates a list of qubit indices involved in a marginal
    unordered_qubits_in_marginal_list = list(reduce(operator.concat, clusters_in_marginal_list))
    
    #creates an inverse noise matrix for the marginal 
    mitigated_clusters_marginal_probability_distribution = math.permute_composite_vector(unordered_qubits_in_marginal_list, final_marginal)

    #qubits list is sorted

    unordered_qubits_in_marginal_list.sort()

    if ensure_proper_probability_distribution:
        if not probability.is_valid_probability_vector(mitigated_clusters_marginal_probability_distribution):
           mitigated_clusters_marginal_probability_distribution = povmtools.find_closest_prob_vector_l2(mitigated_clusters_marginal_probability_distribution)

    mitigated_marginal_probability_distribution = probability.compute_marginal_of_probability_distribution(mitigated_clusters_marginal_probability_distribution ,[unordered_qubits_in_marginal_list.index(qubit) for qubit in marginal])


 

        

    #returns marginalization of an  extended marginal


    return mitigated_marginal_probability_distribution.flatten()

def mitigate_marginals_product(marginals_list: List[Tuple], results_dictionary: Dict[str, Dict[str, int]], noise_model: type[CNNoiseModel] ,ensure_proper_probability_distribution = False )->Dict[str, Dict[tuple[int], np.array]]:
    
    #JT this dictionary 
    mitigated_marginals_dictionary ={}
   
    for marginal in marginals_list:
        mitigated_marginals_dictionary[marginal] = mitigate_marginal_product(marginal=marginal,results_dictionary=results_dictionary,noise_model=noise_model,ensure_proper_probability_distribution=ensure_proper_probability_distribution)
   
    return {next(iter(results_dictionary)): mitigated_marginals_dictionary}
